/*******************************************************************************
 * Copyright (c) 2012 Olivier Langella <Olivier.Langella@moulon.inra.fr>.
 * 
 * This file is part of XTandemPipeline.
 * 
 *     XTandemPipeline is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 * 
 *     XTandemPipeline is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with XTandemPipeline.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contributors:
 *     Benoit Valot <Benoit.Valot@moulon.inra.fr>
 *     Olivier Langella <Olivier.Langella@moulon.inra.fr>
 ******************************************************************************/

package fr.inra.pappso.xtandempipeline.sax_parse;

import java.util.ArrayList;
import java.util.Hashtable;

import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

import org.apache.log4j.Logger;
import fr.inra.moulon.proticport.model.exceptions.ProticPortException;
import fr.inra.pappso.xtandempipeline.class_msms.Identification;
import fr.inra.pappso.xtandempipeline.class_msms.Modifs;
import fr.inra.pappso.xtandempipeline.class_msms.MsRun;
import fr.inra.pappso.xtandempipeline.class_msms.MsRunSet;
import fr.inra.pappso.xtandempipeline.class_msms.Peptide;
import fr.inra.pappso.xtandempipeline.class_msms.Protein;
import fr.inra.pappso.xtandempipeline.grouping.Group;
import fr.inra.pappso.xtandempipeline.grouping.HashSampleScanSetProt;
import fr.inra.pappso.xtandempipeline.grouping.SubGroup;

public class ProticDbMlWriter {
	private static final Logger logger = Logger
			.getLogger(ProticDbMlWriter.class);

	static String namespaceURI = "http://sashimi.sourceforge.net/schema_revision/mzXML_2.0";
	static String xmlnsxsi = "http://www.w3.org/2001/XMLSchema-instance";
	static String xsischemaLocation = "http://uln.moulon.inra.fr/~langella/proticdev/doc/xml_schema/proticdbml/proticdbml.xsd";

	protected XMLStreamWriter writer;

	private ArrayList<Identification> arrayIdentifications = new ArrayList<Identification>();

	private Hashtable<String, String> prot_to_id = new Hashtable<String, String>();

	private Hashtable<String, String> pep_to_id = new Hashtable<String, String>();

	private Hashtable<String, String> sample_to_id = new Hashtable<String, String>();

	private Hashtable<String, String> msrun_to_id = new Hashtable<String, String>();

	private long duracel;

	ProticDbMlWriter(XMLStreamWriter writer) throws ProticPortException {
		this.writer = writer;

		try {
			this.initialize();
		} catch (XMLStreamException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	protected void initialize() throws XMLStreamException, ProticPortException {
		logger.debug("write begin");
		duracel = System.currentTimeMillis();
		// xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
		writer.setPrefix("xsi", xmlnsxsi);
		// writer.setDefaultNamespace(namespaceURI);
		writer.writeStartElement("PROTICdb");
		writer.writeNamespace("xsi", xmlnsxsi);
		// writer.writeDefaultNamespace(namespaceURI);

		// writer.writeAttribute(xmlnsxsi, "schemaLocation", xsischemaLocation);
		//writer.writeAttribute("xmlns", "xsi", xmlnsxsi);

		writer.writeAttribute(xmlnsxsi, "noNamespaceSchemaLocation",
				xsischemaLocation);

		writer.writeAttribute("version", "1.0");

		writer.writeAttribute("type", "MSidentificationResults");

		writeIdentMethod();
		writer.writeStartElement("sequences");

	}

	void write(Identification identification) {
		// protein_list
		logger.debug("write begin");
		try {

			this.arrayIdentifications.add(identification);

			// this.sequences = doc.createElement("sequences");
			this.writeSequence(identification);

		} catch (Exception e) {
			logger.error(e.getMessage());
		}
		logger.debug("write end");
	}

	public void close() throws Exception {
		writer.writeEndElement();// sequences

		writer.writeStartElement("project");
		writer.writeAttribute("name", "");
		writer.writeAttribute("id", "p1");
		// ajout des echantillons_msrun
		this.writeSamples();
		this.writeMsRuns();

		writer.writeStartElement("identificationRuns");

		for (Identification identification : this.arrayIdentifications) {
			writer.writeStartElement("identificationRun");
			// ajout d'une mnip d'identification
			this.writeIdentificationRun();

			// ajout des petides
			this.writepeptideHits(identification);

			// ajout des mathces
			this.writeMatchs(identification);

			writer.writeEndElement();// identificationRun
		}
		writer.writeEndElement();// identificationRuns
		writer.writeEndElement();// project

		writer.writeEndDocument();

		writer.flush();
		writer.close();

		logger.info("Duration creating PROTICdbML xml document: "
				+ (System.currentTimeMillis() - duracel) + " ms");
	}

	private void writeIdentMethod() {
		logger.debug("write begin");
		try {
			// Ajout des méthodes
			writer.writeStartElement("identMeth");
			writer.writeAttribute("name", "");
			writer.writeAttribute("id", "m1");
			writer.writeEndElement();

			// Ajout des base de données
			writer.writeStartElement("customDb");
			writer.writeAttribute("name", "");
			writer.writeAttribute("id", "customdb0");
			writer.writeEndElement();

		} catch (Exception e) {
			logger.error(e.getMessage());
		}
		logger.debug("write end");

	}

	private void writeSamples() throws Exception {
		try {
			writer.writeStartElement("samples");

			for (Identification ident : arrayIdentifications) {
				for (MsRun samp : ident.get_samples()) {
					writer.writeStartElement("sample");
					String id_samp = "samp" + sample_to_id.size();
					String id_msrun = "msr" + msrun_to_id.size();
					String name = samp.getSampleName();
					name = name.replaceAll(".xml", "");
					// balise sample
					writer.writeAttribute("name", name);
					writer.writeAttribute("id", id_samp);
					writer.writeEmptyElement("description");
					sample_to_id.put(samp.getSampleName(), id_samp);

					// Element spectrumList = doc.createElement("spectrumList");
					// msRun.appendChild(spectrumList);

					msrun_to_id.put(samp.getSampleName(), id_msrun);
					writer.writeEndElement();// sample
				}
			}
			writer.writeEndElement();// samples
		} catch (Exception e) {
			String message = "error writing samples in ProticDbMl \n"
					+ e.getMessage();
			logger.error(message);
			throw new Exception(message);
		}
	}

	private void writeMsRuns() throws Exception {
		try {
			writer.writeStartElement("msRuns");

			for (Identification ident : arrayIdentifications) {
				for (MsRun samp : ident.get_samples()) {

					String name = samp.getSampleName();
					name = name.replaceAll(".xml", "");

					String id_samp = sample_to_id.get(samp.getSampleName());

					// Element spectrumList = doc.createElement("spectrumList");
					// msRun.appendChild(spectrumList);

					String id_msrun = msrun_to_id.get(samp.getSampleName());

					// balise MsRun
					writer.writeStartElement("msRun");
					writer.writeAttribute("sample_id", id_samp);
					writer.writeAttribute("id", id_msrun);

					writer.writeStartElement("description");
					writer.writeStartElement("admin");
					writer.writeStartElement("contact");
					writer.writeStartElement("email");
					writer.writeCharacters("valot@moulon.inra.fr");
					writer.writeEndElement();// email

					writer.writeStartElement("name");
					writer.writeCharacters("Valot Benoit");
					writer.writeEndElement();// name
					writer.writeEndElement();// contact

					writer.writeStartElement("sourceFile");
					writer.writeStartElement("nameOfFile");
					writer.writeCharacters(name + ".RAW");
					writer.writeEndElement();// nameOfFile
					writer.writeEndElement();// sourceFile

					writer.writeStartElement("sampleName");
					writer.writeCharacters(name);
					writer.writeEndElement();// sampleName
					writer.writeEndElement();// admin
					writer.writeEndElement();// description

					writer.writeEndElement();// msRun
				}
			}
			writer.writeEndElement();// msRuns
		} catch (Exception e) {
			String message = "error writing MS runs in ProticDbMl \n"
					+ e.getMessage();
			logger.error(message);
			throw new Exception(message);
		}
	}

	private void writeSequence(Identification ident) throws Exception {
		try {
			for (Group group : ident.getGrouping().getGroupList()) {
				for (SubGroup sg : group.getSubGroupSet().getSubGroupList()) {
					for (HashSampleScanSetProt hashProt : sg
							.getHashSampleScanSetProtList()) {
						Protein prot = hashProt.getMatch().get_protein_match();
						if (prot_to_id.containsKey(prot.get_description()) == false) {
							String id = "seq" + prot_to_id.size();
							writer.writeStartElement("sequence");
							writer.writeAttribute("id", id);
							if (prot.get_description().length() > 60) {
								writer.writeAttribute("display_id", prot
										.get_description().substring(0, 59));
							} else {
								writer.writeAttribute(
										"display_id",
										prot.get_description()
												.substring(
														0,
														prot.get_description()
																.length()));
							}
							// <dbxref key="AT5G16390"
							// dbname="AGI_LocusCode"></dbxref>
							if (prot.get_dbxref_type().equals("no") == false) {
								writer.writeStartElement("dbxref");
								writer.writeAttribute("key",
										prot.get_dbxref_accession());
								writer.writeAttribute("dbname",
										prot.get_dbxref_type());
								writer.writeEndElement();// dbxref
							}
							writer.writeStartElement("description");
							writer.writeCharacters(prot.get_description());
							writer.writeEndElement();// description
							writer.writeStartElement("seq");
							writer.writeCharacters(prot.get_sequence());
							writer.writeEndElement();// seq

							writer.writeEndElement();// sequence

							prot_to_id.put(prot.get_description(), id);
						}
					}
				}
			}

		} catch (Exception e) {
			String message = "error writing sequences in ProticDbMl \n"
					+ e.getMessage();
			logger.error(message);
			throw new Exception(message);
		}
	}

	private void writeIdentificationRun() throws Exception {
		try {
			writer.writeAttribute("ident_method_id", "m1");
			writer.writeAttribute("customdb_id", "customdb0");

			writer.writeStartElement("description");
			writer.writeStartElement("admin");
			writer.writeStartElement("contact");
			writer.writeStartElement("email");
			writer.writeCharacters("valot@moulon.inra.fr");
			writer.writeEndElement();// email

			writer.writeStartElement("name");
			writer.writeCharacters("Valot Benoit");
			writer.writeEndElement();// name
			writer.writeEndElement();// contact
			writer.writeEndElement();// admin

			writer.writeStartElement("dataProcessing");
			writer.writeStartElement("software");
			writer.writeStartElement("name");
			writer.writeAttribute("acc", "PROTICdbO:0000283");
			writer.writeCharacters("Xtandem");
			writer.writeEndElement();// name
			writer.writeStartElement("version");
			writer.writeCharacters("2008.02.01.3");
			writer.writeEndElement();// version
			writer.writeEndElement();// software
			writer.writeEndElement();// dataProcessing
			writer.writeEndElement();// description
		} catch (Exception e) {
			String message = "error writing identificationRun in ProticDbMl \n"
					+ e.getMessage();
			logger.error(message);
			throw new Exception(message);
		}
	}

	private void writePeptideHit(Peptide pep) throws XMLStreamException {
		String pep_hit_id = "pep" + pep_to_id.size();
		// param par default
		writer.writeStartElement("peptideHit");
		writer.writeAttribute("calc_mr", "" + (pep.get_mhplus_theo() - 1.00794));
		double exp_mz = (pep.get_mhplus_obser() + (1.00794 * (pep.get_charge() - 1)))
				/ pep.get_charge();
		writer.writeAttribute("exp_mz", "" + exp_mz);
		writer.writeAttribute("delta", "" + pep.get_deltamass());
		writer.writeAttribute("exp_mr", "" + (pep.get_mhplus_obser() - 1.00794));
		writer.writeAttribute("acq_number", "" + pep.get_scan());
		writer.writeAttribute("ms_run_id",
				"" + msrun_to_id.get(pep.get_sample()));
		writer.writeAttribute("id", pep_hit_id);
		writer.writeAttribute("exp_z", "" + pep.get_charge());

		// cvparam specifique xtandem
		writer.writeStartElement("cvParam");
		writer.writeAttribute("value", "" + pep.get_evalue());
		writer.writeAttribute("name", "xtandem peptide evalue");
		writer.writeAttribute("cvLabel", "PROTICdbO");
		writer.writeAttribute("accession", "PROTICdbO:0000287");
		writer.writeEndElement();// cvParam

		writer.writeStartElement("cvParam");
		writer.writeAttribute("value", "" + pep.get_hypercorr());
		writer.writeAttribute("name", "xtandem peptide hyperscore");
		writer.writeAttribute("cvLabel", "PROTICdbO");
		writer.writeAttribute("accession", "PROTICdbO:0000288");
		writer.writeEndElement();// cvParam

		writer.writeStartElement("cvParam");
		writer.writeAttribute("value", "" + pep.get_RT());
		writer.writeAttribute("name", "xtandem peptide retention time");
		writer.writeAttribute("cvLabel", "PROTICdbO");
		writer.writeAttribute("accession", "PROTICdbO:0000289");
		writer.writeEndElement();// cvParam

		// Element cvParam4 = doc.createElement("cvParam");
		// cvParam4.setAttribute("value", "" + pep.get_post());
		// cvParam4.setAttribute("name", "xtandem peptide post");
		// cvParam4.setAttribute("cvLabel", "PROTICdbO");
		// cvParam4.setAttribute("accession", "PROTICdbO:00000XX");
		// peptideHit.appendChild(cvParam4);
		//
		// Element cvParam5 = doc.createElement("cvParam");
		// cvParam5.setAttribute("value", "" + pep.get_pre());
		// cvParam5.setAttribute("name", "xtandem peptide pre");
		// cvParam5.setAttribute("cvLabel", "PROTICdbO");
		// cvParam5.setAttribute("accession", "PROTICdbO:00000XX");
		// peptideHit.appendChild(cvParam5);

		// sequences avec les modifs
		writer.writeStartElement("pepSeq");
		writer.writeStartElement("peptide");
		writer.writeCharacters("" + pep.get_sequence());
		writer.writeEndElement();// peptide
		if (pep.get_Modifs().size() != 0) {
			this.writePtms(pep.get_Modifs());
		}
		writer.writeEndElement();// pepSeq

		pep_to_id.put(pep.get_sample_scan_sequencepeptide_cle(), pep_hit_id);
		writer.writeEndElement();// peptideHit
	}

	private void writePtms(ArrayList<Modifs> mods) throws XMLStreamException {
		writer.writeStartElement("ptms");
		for (Modifs mod : mods) {
			writer.writeStartElement("ptm");
			writer.writeAttribute("diff_mono", "" + mod.get_modvalue());
			writer.writeAttribute("position", "" + mod.get_posi());
			writer.writeAttribute("aa", "" + mod.get_AA());
			this.writeOboModif(mod.get_modvalue());
			writer.writeEndElement();// ptm

		}
		writer.writeEndElement();// ptms
	}

	private void writeOboModif(float F) throws XMLStreamException {
		// <cvParam name="methionine oxidation with neutral loss of 80 Da"
		// cvLabel="PSI-MOD" accession="MOD:00768"></cvParam>
		String name = null;
		String accession = null;
		String cvLabel = null;
		if (((F - 15.99) > -0.01) && ((F - 15.99) < 0.01)) {
			name = "oxidation to L-methionine sulfoxide";
			cvLabel = "PSI-MOD";
			accession = "MOD:00719";
			// oxidation to L-methionine sulfoxide [MOD:00719] 15.994915
		}
		if (((F - 57.02) > -0.03) && ((F - 57.02) < 0.03)) {
			name = "iodoacetamide";
			cvLabel = "PSI-MOD";
			accession = "MOD:00397";
			// iodoacetamide [MOD:00397] 57.021464
		}
		if (((F - 42.01) > -0.01) && ((F - 42.01) < 0.01)) {
			name = "N-acetylated residue";
			cvLabel = "PSI-MOD";
			accession = "MOD:00408";
			// N-acetylated residue [MOD:00408] 42.010565
		}
		if (((F - 79.966) > -0.01) && ((F - 79.966) < 0.01)) {
			name = "phosphorylated residue";
			cvLabel = "PSI-MOD";
			accession = "MOD:00696";
			// phosphorylated residue [MOD:00696] 79.966330
		}
		if (((F + 17.026) > -0.01) && ((F + 17.026) < 0.01)) {
			name = "loss of ammonia";
			cvLabel = "PSI-MOD";
			accession = "MOD:01160";
			// loss of ammonia [MOD:01160] -17.026549
		}
		if (((F + 18.01) > -0.01) && ((F + 18.01) < 0.01)) {
			name = "dehydrated residue";
			cvLabel = "PSI-MOD";
			accession = "MOD:00704";
			// dehydrated residue [MOD:00704] -18.010565
		}
		if (name != null) {
			writer.writeStartElement("cvParam");
			writer.writeAttribute("name", name);
			writer.writeAttribute("cvLabel", cvLabel);
			writer.writeAttribute("accession", accession);
			writer.writeEndElement();// cvParam
		}
		return;
	}

	private void writepeptideHits(Identification identification)
			throws Exception {
		try {
			writer.writeStartElement("peptideHits");
			for (Group group : identification.getGrouping().getGroupList()) {
				for (SubGroup sg : group.getSubGroupSet().getSubGroupList()) {
					for (HashSampleScanSetProt hashProt : sg
							.getHashSampleScanSetProtList()) {
						for (Peptide pep : hashProt.getMatch()
								.get_peptide_order_inclus_redon()) {
							if (pep_to_id.containsKey(pep
									.get_sample_scan_sequencepeptide_cle()) == false) {

								writePeptideHit(pep);

							}
						}
					}
				}
			}
			writer.writeEndElement();// "peptideHits");
		} catch (Exception e) {
			String message = "error writing peptideHits in ProticDbMl \n"
					+ e.getMessage();
			logger.error(message);
			throw new Exception(message);
		}
	}

	private void writeMatchs(Identification identification) throws Exception {
		try {
			writer.writeStartElement("matchs");
			for (Group group : identification.getGrouping().getGroupList()) {
				for (SubGroup sg : group.getSubGroupSet().getSubGroupList()) {
					writer.writeStartElement("match");

					// echantillons
					Hashtable<String, String> view = new Hashtable<String, String>();
					for (HashSampleScanSetProt hashProt : sg
							.getHashSampleScanSetProtList()) {
						for (MsRun sample : hashProt.getMatch()
								.get_sample_matchs()) {
							if (!view.containsKey(sample.getSampleName())) {
								writer.writeStartElement("matchSample");
								writer.writeAttribute(
										"sample_id",
										""
												+ sample_to_id.get(sample
														.getSampleName()));
								String name = sample.getSampleName();
								name = name.replaceAll(".xml", "");
								writer.writeAttribute("name", "" + name);
								view.put(sample.getSampleName(), "ok");
								writer.writeEndElement();// "matchSample");
							}
						}
					}

					// proteins
					int k = 1;
					for (HashSampleScanSetProt hashProt : sg
							.getHashSampleScanSetProtList()) {
						writer.writeStartElement("proteinHit");
						Protein prot = hashProt.getMatch().get_protein_match();
						writer.writeAttribute("sequence_id",
								"" + prot_to_id.get(prot.get_description()));
						// proteinHit.setAttribute("score", "");
						writer.writeAttribute("rank", "" + k);

						// //cvparam
						writer.writeStartElement("cvParam");
						writer.writeAttribute("value", "" + prot.get_mw());
						writer.writeAttribute("name",
								"MW determination by Benuch");
						writer.writeAttribute("cvLabel", "PROTICdbO");
						writer.writeAttribute("accession", "PROTICdbO:0000284");
						writer.writeEndElement();// cvParam

						// evalue
						MsRunSet samples = hashProt.getMatch()
								.get_sample_matchs();
						writer.writeStartElement("cvParam");
						writer.writeAttribute("value", "" + prot.get_evalue());

						writer.writeAttribute("name", "Xtandem log evalue");
						writer.writeAttribute("cvLabel", "PROTICdbO");
						writer.writeAttribute("accession", "PROTICdbO:0000291");
						writer.writeEndElement();// cvParam

						// coverage
						if (samples.size() == 1) {
							writer.writeStartElement("cvParam");

							writer.writeAttribute("value", ""
									+ hashProt.getMatch().get_coverage());
							writer.writeAttribute("name",
									"Coverage determination by Benuch");
							writer.writeAttribute("cvLabel", "PROTICdbO");
							writer.writeAttribute("accession",
									"PROTICdbO:0000285");
							writer.writeEndElement();// cvParam
						}
						writer.writeEndElement();// "proteinHit");
						k++;
					}
					for (HashSampleScanSetProt hashProt : sg
							.getHashSampleScanSetProtList()) {
						Protein prot2 = hashProt.getMatch().get_protein_match();
						// peptidesHitRef
						for (Peptide pep : hashProt.getMatch()
								.get_peptide_order_inclus_redon()) {
							writer.writeStartElement("peptideHitRef");
							writer.writeAttribute(
									"peptide_hit_id",
									""
											+ pep_to_id.get(pep
													.get_sample_scan_sequencepeptide_cle()));

							writer.writeStartElement("fromSeq");
							writer.writeAttribute(
									"seq_id",
									""
											+ prot_to_id.get(prot2
													.get_description()));
							writer.writeAttribute("start", "" + pep.get_start());
							writer.writeAttribute("stop", "" + pep.get_stop());
							String res1 = pep.get_pre();
							res1 = res1.replaceAll("]", "-");
							writer.writeAttribute("residue_before_nter", ""
									+ res1.charAt((res1.length() - 1)));
							String res2 = pep.get_post();
							res2 = res2.replaceAll("]", "-");
							writer.writeAttribute("residue_after_cter", ""
									+ res2.charAt(0));

							writer.writeEndElement();// fromSeq
							writer.writeEndElement();// peptideHitRef
						}

					}
					writer.writeEndElement();// "match");
				}
			}
			writer.writeEndElement();// "matchs");
		} catch (Exception e) {
			String message = "error writing Matchs in ProticDbMl \n"
					+ e.getMessage();
			logger.error(message);
			throw new Exception(message);
		}
	}

}
